/*
      *****************************************************************************
      *  Copyright (C)  by Reinhard Heuberger                                     *
      *                                                                           *
      *  All rights reserved.                                                     *
      *                                                                           *
      *  Redistribution and use in source and binary forms, with or without       *
      *  modification, are permitted provided that the following conditions       *
      *  are met:                                                                 *
      *                                                                           *
      *  1. Redistributions of source code must retain the above copyright        *
      *     notice, this list of conditions and the following disclaimer.         *
      *  2. Redistributions in binary form must reproduce the above copyright     *
      *     notice, this list of conditions and the following disclaimer in the   *
      *     documentation and/or other materials provided with the distribution.  *
      *  3. Neither the name of the author nor the names of its contributors may  *
      *     be used to endorse or promote products derived from this software     *
      *     without specific prior written permission.                            *
      *                                                                           *
      *****************************************************************************
*/  
//
module adress_mux (
        Mode,           // Emulator/clone mode
        CloneWrite,     // Clone mode: Read/Write
        clk_65_6,
        ain_1,          // Adress #1
        cin_1,          // 16bit-Clock #1		  
        ain_0,          // Adress #0
        cin_0,          // 16bit-clock #0
		sclk_0,			//
		sclk_1,         //
        //
        aout,           // Adress out
        cout,           // Clock  out
        CloneWrite_Out,
		sclk_o);
// 
input [12:0] ain_1; 
input [12:0] ain_0; 
input Mode, CloneWrite, clk_65_6;
input cin_0, cin_1;
input sclk_0, sclk_1; 
output [12:0] aout;
output cout, CloneWrite_Out;
output sclk_o;
//
reg [12:0] aout;
reg  cout, CloneWrite_Out, sclk_o; 
//
initial
begin
  aout <= 0;
  cout <= 0;
end
//
always @ (posedge clk_65_6)
begin
    if(Mode) begin                                   // *** clone mode ***
	     //
        if(CloneWrite) begin                         // --- clone Write - mode
            CloneWrite_Out <= 0;                     // Write = 0
            cout <= cin_0;
            aout <= ain_0;
		      sclk_o <= sclk_1;
        end else if(!CloneWrite) begin               // --- clone Write - mode 
            CloneWrite_Out <= 1;                     // Write = 1
            cout <= cin_1;
            aout <= ain_1;                			  //
			   sclk_o <= sclk_0;	
        end
		  //
    end else if(!Mode) begin                         // *** emulator mode ***
        cout <= cin_1;
        aout <= ain_1;
        CloneWrite_Out <= 0;                  		  // Write = 0
		sclk_o <= sclk_0;
    end
end
//
endmodule