/*
      *****************************************************************************
      *  Copyright (C) by Reinhard Heuberger , www.pdp11gy.com                    *
      *                                                                           *
      *  All rights reserved.                                                     *
      *                                                                           *
      *  Redistribution and use in source and binary forms, with or without       *
      *  modification, are permitted provided that the following conditions       *
      *  are met:                                                                 *
      *                                                                           *
      *  1. Redistributions of source code must retain the above copyright        *
      *     notice, this list of conditions and the following disclaimer.         *
      *  2. Redistributions in binary form must reproduce the above copyright     *
      *     notice, this list of conditions and the following disclaimer in the   *
      *     documentation and/or other materials provided with the distribution.  *
      *  3. Neither the name of the author nor the names of its contributors may  *
      *     be used to endorse or promote products derived from this software     *
      *     without specific prior written permission.                            *
      *                                                                           *
      *****************************************************************************
      
      This module is based on a merge of a MFM DECODER and data&adress SEQUENCER
	  
                                     Decoding
    Based on a SoC/HPS environment like the DE10-Nano board: MFM decoder module for the 
    DEC RL01/RL02 disk drives. This decoder is supplied with a  8 times faster clock
    runnning at 65.6 MHz to generate a jitter free 4.1MHz clock. Clock recovery is not
    necessary because the clock is already included within the MFM data sequence.
                                    Sequencing
    This module generates the sector addresses and the addresses within one sector for 
    one track. A track consists of 40 sectors (0-39) and a each sector consists of 140
    16-bit words including servo / header / CRC information. In the FPGA, a sector 
    with 144 16-bit words is implemented for the purpose of 16-bit boundary.                                    
    
                             Version 2.7B4 / JANUAR 2021
                    part of the RL01/RL02 disk reader/emulator Version V2.7
    
    
*/
module      clone_MFM_DEcoder(
            Enable,                             // Enable
            System_clk,                         // RL System_clk, 4.1MHz
            Clk_65_6,                           // System Clock 65.6 MHz
            MFM_in,                             // MFM signal 
            Sector,                             // Start decoding (@ sector)
            //
            Parallel_out,                       // parallel output
            address_out,                        // 13 bit adress out
            Clock_16bit,                        // Clock 16 Bit
            Serial_out,                         // Serial   Output
            start_MFM,                          // First MFM puls in sector
            scan_mfm,                           // MFM to be scanned
            clk_puls,                           // 4.1MHz clock puls
            clk_4_1);                           // RL System_clk, 4.1MHz
            //          
input  Enable, System_clk, Clk_65_6, MFM_in, Sector;
output [15:0]  Parallel_out;
output [12:0]  address_out;
output Clock_16bit,Serial_out, start_MFM; 
output scan_mfm, clk_puls, clk_4_1;
//
reg [15:0]  sync_4_1;
reg [15:0]  sync_mfm;
reg [23:0] shifter0;                            // shifter, serial->parallel
reg [31:0] word_clock;                          // 16 Bit clock
reg [15:0] latch_word;                          // Parallel out
reg [31:0] current_address;                     // augenblickliche Addresse^
reg [31:0] grenze_unten;                        // Start Addresse ( DP RAM )
reg [31:0] grenze_oben;                         //  bis gültige Obergrenze.
reg [31:0] sector_counter;                      // Sector counter
reg [12:0] address_out;                         // 13-Bit output Adress to DPR
reg FlipFlop, CLK_16bit_P; 
reg firstMFM, clksel, Sector_p, sync_p;
wire load_word_n, Sector_puls;
reg found, foundf, check;
wire clock_pulse,decide;
wire first_MFM_pulse;
//
initial
begin
    word_clock       <= 0;
    FlipFlop         <= 0;
    firstMFM         <= 0;
    found            <= 0;
    foundf           <= 0;
    check            <= 0;
    current_address  <= 0;
    grenze_unten     <= 0;
    grenze_oben      <= 140;
end 
//
//
//          
//
//
//--------------------------- Find first MFM pulse --------------------------
//---------------------------------------------------------------------------
always @ (posedge Clk_65_6)
begin
    if(Enable & Sector) begin
        if(!firstMFM) begin
           if(MFM_in) begin
                firstMFM <= 1;            // indicate: first MFM puls
                found    <= 1;            // indicate: first MFM puls
            end else begin
                firstMFM <= 0;
                found    <= 0; 
            end
        end else if(firstMFM & found & !MFM_in ) begin
            found <= 0;
        end else begin
            if(!check) begin
                 foundf <= found & System_clk;
                 check  <= 1;
            end
          end
    end else begin
        firstMFM       <= 0;
        found          <= 0;
        foundf         <= 0;
        check          <= 0;
    end
end
assign decide   = foundf;
assign start_MFM = firstMFM;        
//          
//          
//_-_-_-_-_-_-_-_-_-_-_-_-_ Synchronize 4.1MHz clock  -_-_-_-_-_-_-_-_-_-_-_-_
//-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_
always @ (posedge Clk_65_6)
begin
    sync_4_1 <= { sync_4_1[14:0] , System_clk };
    if(decide) begin
        clksel = (!sync_4_1[4] &  sync_4_1[5]);  // (!sync_4_1[4] &  sync_4_1[5]);
    end else begin
        clksel = (!sync_4_1[5] &  sync_4_1[4]);  // (!sync_4_1[5] &  sync_4_1[4]);
    end
end
assign clk_puls = clksel;
assign clk_4_1  = sync_4_1[4];                           // LA output
//          
//          
//-------------------- Synchronize MFM signal to FPGA clock ------------------
//----------------------------------------------------------------------------
always @ (posedge Clk_65_6)
begin
    sync_mfm <= { sync_mfm[14:0] , MFM_in };               // shift MFM <<<
end
assign scan_mfm = (sync_mfm[2] | sync_mfm[1]);             // (sync_mfm[4] | sync_mfm[3])
assign first_MFM_pulse = (!sync_mfm[1] &  sync_mfm[0]);    // 80Mhz MFM pulse #2
// 
//         
//
//
//
//
//                                +--------------------+
//                                |----- "decode" -----|
//                                +--------------------+
//
always @ (posedge Clk_65_6)
begin
    if(Enable) begin
        if(Sector) begin                            // sector Pulse = High
            //
            if(first_MFM_pulse & word_clock == 0 ) begin
                word_clock <= word_clock - 1;
            end
            if(clk_puls & start_MFM ) begin
                if (scan_mfm) begin
                    FlipFlop <= 1;
                end else begin
                    FlipFlop <= 0;
                end         
                shifter0 <= { shifter0[22:0] , FlipFlop};
                word_clock <= word_clock - 1; 
            end 
            //
        end else begin
            FlipFlop    <= 0;
            word_clock  <= 1;                        // Preset counter(7)
        end
    end else begin
        FlipFlop <= 0;
        shifter0 <= 0;
    end
end
assign Clock_16bit   = word_clock[3];
assign Serial_out    = FlipFlop;
//
//
//=============== Generate load pulse for parallel out =======================
//============================================================================
always @ (posedge Clk_65_6)
begin
  CLK_16bit_P <= Clock_16bit;
end
assign load_word_n = CLK_16bit_P &  !Clock_16bit;              // negativ edge
//
//=========================== latch 16 bit output ============================
//============================================================================
always @ (posedge Clk_65_6)
begin
  if (load_word_n) begin                          // little endian ordering
    latch_word[15] <= shifter0[0]; 
    latch_word[14] <= shifter0[1];
    latch_word[13] <= shifter0[2];
    latch_word[12] <= shifter0[3];
    latch_word[11] <= shifter0[4];
    latch_word[10] <= shifter0[5];
    latch_word[9]  <= shifter0[6];
    latch_word[8]  <= shifter0[7];
    //
    latch_word[7]  <= shifter0[8];  
    latch_word[6]  <= shifter0[9];
    latch_word[5]  <= shifter0[10];
    latch_word[4]  <= shifter0[11];
    latch_word[3]  <= shifter0[12];
    latch_word[2]  <= shifter0[13];
    latch_word[1]  <= shifter0[14];
    latch_word[0]  <= shifter0[15];
  end
end
assign Parallel_out = latch_word;                          // 8bit, = byte  out
//
//
//
//
//
//                                +----------------------+
//                                |----- "sequence" -----|
//                                +----------------------+ 
//
always @ (posedge Clk_65_6)
begin
    Sector_p <= Sector;
end
assign Sector_puls = Sector_p & !Sector;                                  
//
//================= Sector Idex-Adressen bilden, 0-39 ========================
//========== 16Bit Adressen innerhalb eines Sector bilden, 0-139 =============
//============================================================================
always @ (posedge Clk_65_6)
begin
    //if ( ~synced_SECTOR_clk[1] & synced_SECTOR_clk[2] )    // @ negedge SEC clk
    if (Sector_puls) begin
        if(Enable) begin
            sector_counter = sector_counter +1;              // Increment
            if ( sector_counter == 40 ) begin                //  If Limit ( 0-39 )    
                sector_counter  <= 0;                        //      clear counter    
                current_address <= 0;                        //      and address
                grenze_unten    <= 0;
                grenze_oben <= 140;                          //  set upper limit 
            end else if( sector_counter < 40 ) begin
                grenze_unten <= grenze_unten + 144;          //  DP-RAM-Index = +144
                current_address <= grenze_unten;             //  Neue untere Grenze
                grenze_oben <= grenze_unten + 140;           //  set upper limit  
            end    
        end   
    end
    //
    // Synchron zur 50Mhz clock werden nun mit der negativen Flanke der
    // 16Bit clock die Adressen innerhalb eines Sectors gebildet.
    //      
    //if ( ~synced_WORD_clk[1] & synced_WORD_clk[2] )         // @ negedge ENC clk
    if ( load_word_n ) begin                                  // *start*
        if(Enable) begin
            if ( current_address == grenze_oben+2) begin
                 current_address <= current_address;          // nix
            end else begin
                current_address <= current_address + 1;       // words/sector-counter
                address_out <= current_address[12:0];         // send Adress              
            end
        end else begin
            current_address <= 0;
            grenze_unten    <= 0;
            grenze_oben     <= 140;
        end
    end    
end
//
endmodule
