/*
      *****************************************************************************
      *  Copyright (C) by Reinhard Heuberger                                 *
      *                                                                           *
      *  All rights reserved.                                                     *
      *                                                                           *
      *  Redistribution and use in source and binary forms, with or without       *
      *  modification, are permitted provided that the following conditions       *
      *  are met:                                                                 *
      *                                                                           *
      *  1. Redistributions of source code must retain the above copyright        *
      *     notice, this list of conditions and the following disclaimer.         *
      *  2. Redistributions in binary form must reproduce the above copyright     *
      *     notice, this list of conditions and the following disclaimer in the   *
      *     documentation and/or other materials provided with the distribution.  *
      *  3. Neither the name of the author nor the names of its contributors may  *
      *     be used to endorse or promote products derived from this software     *
      *     without specific prior written permission.                            *
      *                                                                           *
      *****************************************************************************


    schift >> 15............2...1...0
    serial out                  |   +---> Marker-flipflop
                                +-------> 0 = "Seek" mode 
                                          1 = "Send Drive Status" mode
                                          
                                         Version 3.0
                    part of the RL01/RL02 disk reader/emulator Version V2.7
*/
//
//
//
// 
//
// 
module clone_command_status(
        Enable,                     // Enable
        clk_65_6,                   // System Clock 65.6 MHz
        Clone_command,              // Data, 16 flipflop parallel in, send command
        Drive_Status,               // Serial in, receive drive Status
        System_clk,                 // RL system Clock, 4.1MHz
        Status_clk,                 // RL status Clock, 4.1MHz
        trigger_command,            // trigger, data set @ PIO
        antwort,
        //
        Clone_status,               // Data, 16 flipflop parallel out, status
        Drive_Command);             // Data, Serial out
//      
input  [15:0] Clone_command;    
input  Enable, clk_65_6, Drive_Status, System_clk; 
input  Status_clk, trigger_command, antwort;
output [15:0] Clone_status;
output Drive_Command;
//
reg [31:0] count1;                  //
reg [31:0] count2;                  // 
reg [16:0] command_rl;              // 16  flipflop latched driv command
reg [16:0] status_rl;               // 16  flipflop latched driv status
reg [7:0]  statusclk;               //
reg [7:0]  systemclk;               //
reg [7:0]  status_in;               // like 2 D-getstatusflag register,
reg trigger_command_p;
reg getstatusflag;
//
wire Status_Clock_puls1;
wire System_Clock_puls1, System_Clock_puls2;
wire synced_status;
wire trigger_puls;
wire get_status;
//
initial
begin
   command_rl     <= 0;
   count1         <= 0;
   count2         <= 0;
end
//
//
//===================== Generate a 20ns trigger pulse ========================
//============================================================================
always @ (posedge clk_65_6)
begin
 trigger_command_p <= trigger_command;
end
assign trigger_puls = !trigger_command_p & trigger_command;
//
//
//----------------- Synchronize Status IN to FPGA clock -----------------------
//----------------------------------------------------------------------------
always @ (posedge clk_65_6)
begin
  status_in <= { status_in[6:0] , Drive_Status };
end
assign synced_status =  status_in[2];
//
//
//----------------- Synchronize Status-4.1MHz clock to FPGA clock ------------
//----------------------------------------------------------------------------
always @ (posedge clk_65_6)
begin
  statusclk <= { statusclk[6:0] , Status_clk };
end
assign Status_Clock_puls1 = (!statusclk[4] & statusclk[3]);    // pulse #1
//
//
//----------------- Synchronize System-4.1MHz clock to FPGA clock ------------
//----------------------------------------------------------------------------
always @ (posedge clk_65_6)
begin
  systemclk <= { systemclk[6:0] , System_clk };
end
assign System_Clock_puls1 = (!systemclk[4] & systemclk[3]);    // pulse #1
assign System_Clock_puls2 = (!systemclk[5] & systemclk[4]);    // pulse #2
//
//
//
//
//
//          send command: parallel in from FPGA=RL_system
//                        serial out to RL02 hard disk drive
//
always @ (posedge clk_65_6)
begin   
    if( Enable ) begin
        if(trigger_puls) begin                                 // Start to send command
            command_rl[16:1]  <= Clone_command[15:0];          // pick up command from PIO
                command_rl[0]     <= 1'b0;                     // Preset dummy 
                count1            <= 0;                        // preset counter
            if(Clone_command[1]) begin                         // get Status?
                getstatusflag <= 1'b1;                         // Status request mode
            end else begin
                getstatusflag <= 1'b0;                         // Command mode
            end
        end 
        if(count1 <= 16) begin                                 // 
            if ( System_Clock_puls1 ) begin                    // pos. edge-
                command_rl <= command_rl >> 1;                 // shift
            end else if ( System_Clock_puls2 ) begin
                command_rl[16] <= 1'b1;                        // & set shifter_in
                count1 <= count1 +1;
            end else begin
                count1 <= count1;
            end
        end else begin
            command_rl[0] <= 1'b0;
        end
    end else begin 
        command_rl    <= 0;                                    // if not enable
    end
end
assign Drive_Command = command_rl[0];
assign get_status    = getstatusflag;
//
//
//
//
//
//          receive status: serial in from RL02 hard disk drive 
//                          parallel out to FPGA=RL_system
//
always @ (posedge clk_65_6)
begin
    if(Enable & !antwort) begin  
        if(get_status) begin 
            if(count2 <= 15) begin          
                if ( Status_Clock_puls1 ) begin                 // posedge 4.1 MHz
                    status_rl <= status_rl >> 1;                // shift
                    status_rl[16] <= synced_status;             // & set shifter_in
                    count2 <= count2 +1;
                end else begin
                    status_rl <=status_rl;
                end
            end else begin
                count2 <= count2;
            end
        end else begin
            count2    <= 0;
        end
    end else begin
        status_rl     <= 0;  
    end   
end
assign Clone_status[15:0] = status_rl[16:1];
//assign Clone_status = status_rl;
//
endmodule
